/*
 *  connect4.cpp
 *
 *  Plays the game of Connect Four
 *
 *  Ken Reneris     12/06/2002
 *
 */


#include "std.h"

//#define EMULATOR


//
// Global images
//

#include "white.bmp.cpp"
#include "black.bmp.cpp"

//
// Defines for the game board
//

#define BRD_X_INDENT    12
#define BRD_X_INCR      24
#define BRD_X_MAX       (BRD_X_INDENT+BRD_X_INCR*7)

#define BRD_Y_INDENT    12
#define BRD_Y_INCR      24
#define BRD_Y_MAX       (BRD_Y_INDENT+BRD_Y_INCR*7)


#define COLOR_GRAY      COLOR_GRAY13

#define FRAME_TIME  (30 * 1000)

//
//
//

C4App *CApp;          // The one and only App


///////////////////////////////////////////////////////////////////////////////////////////
//
//  Global functions
//


S32 GUI_main( MSG_TYPE type, CViewable *object, S32 data )
{

    switch (type) {
    case MSG_APP_START:
        CApp = new C4App();
        break;

    case MSG_APP_STOP:
        break;

    case MSG_MENU_SELECT:
        switch((U16) data) {

        case FMENU_SETTINGS:
            new CRevSettings();
            break;

        case FMENU_NEW_GAME:
            CApp->NewGame();
            break;

        case FMENU_UNDO_MOVE:
            CApp->UndoMove();
            break;

        case FMENU_ABOUT:
            GUI_TextWindow(ID_ABOUT, "Connect Four 1.0",
                "Written By: Ken Reneris\n"
                "Dec 6, 2002\n"
                "\n"
                "Enjoy!"
            );
            break;

        case FMENU_EXIT:
            GUI_Exit();
            break;
        }

    default:
        // unhandled
        return 0;
    }


    return 1;
}





//////////////////////////////////////////////////////////////////////////////////////////
//
// CRevApp
//


C4App::C4App()
    : CWindow(MAIN_WINDOW_ID, 0, 0, 200, 240, "Connect Four")

{
    m_StatLn1 = new CLabel(0, "");
    m_StatLn2 = new CLabel(0, "");
    m_Bar = new CProgressBar(0, BRD_X_MAX);

    // Allocate file menu items
    m_FMenu = new CMenu(FMENU_ID);
    m_FMenu->SetNumRows(7);
    m_FMenu->SetRow(0, FMENU_NEW_GAME,  "New Game",  'n' );
    m_FMenu->SetRow(1, FMENU_UNDO_MOVE, "Undo Move", 'u' );
    m_FMenu->SetRow(2, FMENU_SETTINGS,  "Settings",  's' );
    m_FMenu->SetSeparatorRow(3);
    m_FMenu->SetRow(4, FMENU_ABOUT, "About");
    m_FMenu->SetSeparatorRow(5);
    m_FMenu->SetRow(6, FMENU_EXIT, "Exit");

    m_FMenuUndo = 1;
    m_FMenu->DisableRow(m_FMenuUndo);

    // Put file menu at lower left of display
    m_FMenu->SetY ( GetHeight() - m_FMenu->GetHeight());

    //
    AddChild(m_StatLn1, 0, BRD_Y_MAX + 4);
    AddChild(m_StatLn2, 0, 240 - 30);
    AddChild(m_Bar, BRD_X_INDENT, 240-30);

    //
    // Set default settings
    //

    memset(&m_SetName, 0, sizeof(m_SetName));
    strcpy(m_SetName.publisher, "Ken Reneris");
    strcpy(m_SetName.name, "Game Settings");
    strcpy(m_SetName.extension, "bin");

    memset(&m_Set, 0, sizeof(m_Set));
    m_Set.Sig = SETTINGS_SIG;
    m_Set.Animate = TRUE;
    m_Set.Sound = TRUE;
    m_Set.Depth[R_WHITE] = 4;           // Med computer player
    m_Set.Depth[R_BLACK] = 0;           // Human player
    m_Thinking = FALSE;
    LoadSettings();

    srand(m_Set.Seed);
    m_Set.Seed = rand();
    SaveSettings();

    //
    // Setup for new game
    //

    NewGame();
}

VOID C4App::LoadSettings()
{
    INTN        err;
    size_t      size;
    PCHAR       p;

    // Map in and read our settings object

    p = (PCHAR) OS_availaddr;
    size = ebo_roundup (sizeof(m_Set));
    err  = ebo_mapin(&m_SetName, 0, p, &size, FALSE);
    if (!err) {
        if (((UINTN *) (p+SETTINGS_OFFSET))[0] == SETTINGS_SIG) {
            memcpy(&m_Set, p+SETTINGS_OFFSET, sizeof(m_Set));
        }
        ebo_unmap(p, size);
    }
}


VOID C4App::SaveSettings()
{
    INTN        err;
    size_t      size;
    PCHAR       p;

    // Map in and write our settings object

    p = (PCHAR) OS_availaddr;
    size = ebo_roundup (sizeof(m_Set));
    err  = ebo_mapin(&m_SetName, 0, p, &size, TRUE);

    // If couldn't open, try creating
    if (err) {
        ebo_new(&m_SetName, size, TRUE);
        err = ebo_mapin(&m_SetName, 0, p, &size, TRUE);
    }

    // If opened, write it
    if (!err) {
        memcpy(p+SETTINGS_OFFSET, &m_Set, sizeof(m_Set));
        ebo_unmap(p, size);
    }
}


VOID C4App::NewGame()
{
    m_Brd.InitialBoard();
    m_HistIndex = 0;
    m_GameOver = FALSE;
    m_WinRun = 0;
    m_FMenu->DisableRow(m_FMenuUndo);
    CheckNextMove();
    Draw();
}


VOID C4App::UndoMove()
{
    if (m_HistIndex) {

        // Restore previous history
        m_HistIndex -= 1;
        m_Brd = m_Hist[m_HistIndex];

        // Reset move state
        m_WinRun = 0;
        m_GameOver = FALSE;
        if (!m_HistIndex) {
            m_FMenu->DisableRow(m_FMenuUndo);
        }

        // Done, update game board and verify the computers turn
        CheckNextMove();
        Draw();
    }
}

S32 C4App::MsgHandler(MSG_TYPE type, CViewable *object, S32 data)
{
    INTN        x, y, cx, cy;


    switch (type) {
    case MSG_KEY:
        if (m_WinFlash) {
            m_WinFlash = 0;
            Draw();
        }
        m_FMenu->Show();
        break;

    case MSG_BUTTON_SELECT:
        // Special case where we send ourselves messages in data
        if (data > 1) {
            switch(data) {
            case MAIN_CHECK_NEXT_MOVE:
                ICheckNextMove();
                break;
            case MAIN_TIMER:
                Timer();
                break;
            }
            return 1;
        }
        return 0;
        break;

    case MSG_PEN_DOWN:
        x = (data >> 16)  & 0xFFFF;
        y = (data) & 0xFFFF;
        if (x >= BRD_X_INDENT && x <= BRD_X_MAX &&
            y >= BRD_Y_INDENT && y <= BRD_Y_MAX) {

            // Waiting for human player?
            if (!m_Thinking && IsHumanTurn()) {

                // Covert X,Y to cell
                cx = (x - BRD_X_INDENT) / BRD_X_INCR;
                cy = (y - BRD_Y_INDENT) / BRD_Y_INCR;

                // Can we make this move?
                if (m_Brd.CanMove(cx) && !m_GameOver) {

                    // Yup, do it
                    MakeMove(cx);

                } else {

                    // Nope,
                    PlaySound(SndBadMove);
                }
            }
        }

        break;

    case MSG_TIMER:
        // Send non-interrupt message to process
        GUI_EventMessage(MSG_BUTTON_SELECT, this, MAIN_TIMER);
        break;

    default:
        return 0;
    }
    return 1;
}


VOID C4App::MakeMove(UINTN Col)
{

    //
    // If this is a human, add to undo buffer
    //

    if (IsHumanTurn()) {
        m_Hist[m_HistIndex] = m_Brd;
        m_HistIndex += 1;
        if (m_HistIndex == 1) {
            m_FMenu->EnableRow(m_FMenuUndo);
        }
    }

    //
    // Apply the move to the board
    //

    if (m_Set.Animate) {
        INTN        x, y, s, MaxY;
        IMAGE       *img;
        COLOR       bg;

        bg =  GetBackgroundColor();
        img = (IMAGE *) (m_Brd.m_Color == R_WHITE ? &white_bmp : &black_bmp);

        s = 1;
        x = BRD_X_INDENT + 2 + BRD_X_INCR * Col;
        y = BRD_Y_INDENT + 2;
        MaxY = BRD_Y_INDENT + 2 + BRD_Y_INCR * m_Brd.ColY(Col);

        DrawRectFilled(0, BRD_Y_MAX+2, 200, 240-BRD_Y_MAX-2, bg);
        while (y < MaxY) {
            DrawImage(x, y, img);
            DrawLine(BRD_X_INDENT, BRD_Y_INDENT, BRD_X_MAX, BRD_Y_INDENT);
            GUI_UpdateNow();
            DrawRectFilled(x, y, 21, 21, bg);
            usleep(FRAME_TIME);
            y += s;
            s += 1;
        }

        DrawImage(x, MaxY, img);
        DrawLine(BRD_X_INDENT, BRD_Y_INDENT, BRD_X_MAX, BRD_Y_INDENT);
        GUI_UpdateNow();
    }


    m_Brd.DropChip(Col);
    PlayClick();

    //
    // Switch players
    //

    m_Brd.NextPlayer();
    Draw();

    //
    // Check that it's not our turn
    //

    CheckNextMove();
}


VOID C4App::ICheckNextMove()
{
    UINTN       Col;

    //
    // Don't allow to recurs (as we check for messages during our thinking process)
    //

    if (m_Thinking) {
        return ;
    }

    m_Thinking = TRUE;

    //
    // Is there a winner?
    //

    if (m_Brd.GameOver()) {
        GameOver();

    } else {

        if (IsComputerTurn()) {
            m_Bar->SetPosition(0);
            m_Bar->SetMaximum(7);

            Col = m_Brd.RootMoveSearch(m_Set.Depth[m_Brd.m_Color]);
            MakeMove(Col);
        }
    }

    m_Thinking = FALSE;
}


VOID C4App::CheckNextMove()
{
    GUI_EventMessage(MSG_BUTTON_SELECT, this, MAIN_CHECK_NEXT_MOVE);
}


VOID C4App::DrawProgressBar(INTN Count)
{
    m_Bar->SetPosition(Count);
    m_Bar->Draw();
    GUI_UpdateNow();
    GUI_EventLoop(0);   // Let messages get processed
}


VOID C4App::Timer()
{
    if (m_WinFlash && m_WinRun) {
        m_WinFlash -= 1;
        SetTimer(500, FALSE, 0);
        Draw();
    }
}


VOID C4App::Draw()
// Redraw entire window
{
    RECT        rc;
    INTN        x, y, i, c;
    COLOR       bg;

    bg = GetBackgroundColor();
    GetUsableRect(&rc);

    DrawRectFilled(0,0, rc.width, rc.height, bg);
    DrawTitle();

    //
    // Draw board grid
    //

    for (x=BRD_X_INDENT;  x <= BRD_X_MAX; x += BRD_X_INCR) {
        DrawLine(x, BRD_Y_INDENT, x, BRD_Y_MAX+1);
    }

    DrawLine(BRD_X_INDENT, BRD_Y_INDENT, BRD_X_MAX, BRD_Y_INDENT);
    DrawLine(BRD_X_INDENT, BRD_Y_MAX,    BRD_X_MAX, BRD_Y_MAX);

    // Add some shading
    DrawLine(BRD_X_MAX+1, BRD_Y_INDENT+1, BRD_X_MAX+1, BRD_Y_MAX+2, COLOR_GRAY);
    DrawLine(BRD_X_INDENT, BRD_Y_MAX+1, BRD_X_MAX+1, BRD_Y_MAX+1, COLOR_GRAY);

    //
    // Draw the non-empty squares
    //

    for(x=0; x<R_BOARD_SIZE; x++) {
        for(y=0; y<R_BOARD_SIZE; y++) {
            c = m_Brd.Get(x,y);

            if (c != R_EMPTY) {
                DrawImage(
                    BRD_X_INDENT+2 + x * BRD_X_INCR,
                    BRD_Y_INDENT+2 + y * BRD_Y_INCR,
                    c == R_BLACK ?  &black_bmp : &white_bmp
                    );
            }
        }
    }

    //
    // Mark winners
    //

    if (m_WinFlash & 1) {
        for(i=0; i<m_WinRun; i++) {
            InvertRect(m_Win[i].x, m_Win[i].y, 23, 23);
        }
    }

    DrawTurn();
    ResetUsableRect();
    GUI_UpdateNow();
}


VOID C4App::DrawTurn()
{
    COLOR       bg;
    LPCSTR      l1;
    INTN        w;

    bg = GetBackgroundColor();
    DrawRectFilled(0, BRD_Y_MAX+2, 200, 240-BRD_Y_MAX-2, bg);

    if (!m_GameOver) {
        l1 = m_Brd.m_Color == R_WHITE ? "White Move" : "Black Move";
        if (m_Set.Depth[m_Brd.m_Color]) {
            l1 = m_Brd.m_Color == R_WHITE ? "White Thinking" : "Black Thinking";
        }

        w = GUI_TextWidth(GetFont(), l1, -1);
        DrawText(l1,  (200 - w) / 2, BRD_Y_MAX + 4);

    } else {

        // Draw winner status
        m_StatLn1->Draw();
        m_StatLn2->Draw();

    }
    GUI_UpdateNow();
}


VOID C4App::FindWin(UINTN Color)
{
    CPoint      pt, d;

    for(pt.y=0; pt.y<R_BOARD_SIZE; pt.y++) {
        for(pt.x=0; pt.x<R_BOARD_SIZE; pt.x++) {
            for(d.x=-1; d.x<=1; d.x++) {
                for(d.y=-1; d.y<=1; d.y++) {
                    if (d.x || d.y) {
                        CheckRun(pt,d,Color);
                    }
                }
            }
        }
    }

    m_WinFlash = 8;
    SetTimer(500, FALSE, 0);
}

VOID C4App::CheckRun(CPoint pt, CPoint d, UINTN Color)
{
    INTN        i, j;

    for(i=0; i<4; i++) {
        if (m_Brd.Get(pt.x,pt.y) != (INTN) Color) {
            return ;
        }
        pt += d;
    }

    for(i=0; i<4; i++) {
        pt -= d;
        m_Win[m_WinRun].x = BRD_X_INDENT+1 + pt.x * BRD_X_INCR;
        m_Win[m_WinRun].y = BRD_Y_INDENT+1 + pt.y * BRD_Y_INCR;

        // Remove dups
        for(j=0; j<m_WinRun; j++) {
            if (m_Win[j] == m_Win[m_WinRun]) {
                m_WinRun -=1;
            }
        }

        m_WinRun += 1;
    }
}


VOID C4App::GameOver()
{
    const piezo_tone   *Snd;
    UINTN           Winner;
    LPCSTR          l1;
    BOOLEAN         f;

    if (!m_GameOver) {
        Snd = SndGameOver;
        m_GameOver = TRUE;
        Winner = m_Brd.Winner();

        f = (m_Set.Depth[R_WHITE] == 0 && m_Set.Depth[R_BLACK] == 0) ||
            (m_Set.Depth[R_WHITE] > 0  && m_Set.Depth[R_BLACK] > 0);

        strcpy(m_Ln1, "Tie Game!");
        if (Winner != R_EMPTY) {
            FindWin(Winner);

            if (f) {
                l1 = (Winner == R_WHITE) ? "White wins" : "Black wins";
            } else {
                l1 = "eBookMan wins";
                if (m_Set.Depth[Winner] == 0) {
                    l1 = "You win";
                    Snd = SndYouWin;
                }
            }

            sprintf(m_Ln1, "%s!", l1);
        }

        m_StatLn1->SetLabel(m_Ln1);
        m_StatLn1->SetFont(GUI_GetFont(12, CTRL_BOLD));
        m_StatLn1->SetX( (200 - m_StatLn1->GetWidth()) / 2 );

        // Set m_Ln2 label here
        m_StatLn2->SetLabel(m_Ln2);
        m_StatLn2->SetX( (200 - m_StatLn2->GetWidth()) / 2 );

        // Done - draw it
        PlaySound(Snd);
        DrawTurn();
    }
}

VOID C4App::PlayClick()
{
    if (m_Set.Sound) {
        piezo_play_click(0);
    }
}

VOID C4App::PlaySound(const piezo_tone *Snd)
{
    const piezo_tone *p;
    UINTN       NoNotes;

    if (m_Set.Sound) {
        // Get # of notes to play
        for(p=Snd, NoNotes=0; p->man; p++, NoNotes++) ;

#ifndef EMULATOR
        // Play them
        piezo_play_1jingle(NoNotes, (piezo_tone *) Snd, 0, 0);
#endif
    }
}
